@@ -3,12 +3,13 @@ package BoardStreams::Transaction;
 use Mojo::Base -base, -signatures;
 
 use Mojo::JSON 'from_json', 'to_json', 'encode_json';
+use Mojo::IOLoop;
 use Struct::Diff 'diff';
 use List::AllUtils 'each_array', 'indexes';
 use Storable 'dclone';
 use Carp 'croak';
 
-our $VERSION = "v0.0.17";
+our $VERSION = "v0.0.18";
 
 has bs_prefix => sub { die "bs_prefix is required" };
 has c => sub { die "c is required" };
@@ -57,7 +58,7 @@ sub lock_state ($self, $channel_names, $sub, %opts) {
         { for => 'update' },
     )->hashes;
 
-    if (! $opts{no_ban}) {
+    if (! $opts{no_ban} and Mojo::IOLoop->is_running) {
         my $workers_state = $self->c->$bs_prefix->get_state($WORKERS_CHANNEL);
         $workers_state->{$worker_uuid}
             and ! $workers_state->{$worker_uuid}{banned}
