@@ -6,7 +6,7 @@ use BoardStreams::Registry;
 use BoardStreams::ListenerObservable 'get_listener_observable';
 use BoardStreams::Transaction;
 use BoardStreams::Exception 'db_duplicate_error';
-use BoardStreams::Util 'string', ':bool';
+use BoardStreams::Util 'string', ':bool', 'next_tick_p';
 use BoardStreams::DBMigrations;
 
 use Mojo::Pg;
@@ -25,7 +25,7 @@ use Encode 'encode_utf8';
 
 use experimental 'postderef';
 
-our $VERSION = "v0.0.17";
+our $VERSION = "v0.0.18";
 
 use constant {
     DEFAULT_CLEANUP_INTERVAL          => 5,
@@ -196,24 +196,34 @@ sub register ($self, $app, $config) {
 
         # one process needs to call "global_cleanup" every "cleanup_interval" seconds:
         if ($cleanup_interval) {
-            rx_timer(rand($cleanup_interval), $cleanup_interval + 0.1 + rand())->subscribe(sub {
+            rx_defer(sub {
+                return rx_timer($cleanup_interval * 0.5 + rand($cleanup_interval));
+            })->pipe(
+                op_repeat(),
+            )->subscribe(async sub {
                 my $t0 = Time::HiRes::time();
-                return unless
-                    $t0 >= $app->$bs_prefix->get_state($CLEANUP_CHANNEL)->{last_cleanup} + $cleanup_interval;
+                my $cleanup_state = $app->$bs_prefix->get_state($CLEANUP_CHANNEL);
+                return unless $t0 >= $cleanup_state->{last_cleanup} + $cleanup_interval;
 
                 my $me_is_cleanup_executor = 0;
-                $app->$bs_prefix->lock_state($CLEANUP_CHANNEL, sub ($state) {
-                    if (Time::HiRes::time() >= $state->{last_cleanup} + $cleanup_interval) {
-                        $state->{last_cleanup} = $t0;
-                        $me_is_cleanup_executor = 1;
-                        return undef, $state;
-                    } else {
-                        return undef, undef;
-                    }
+
+                my $txn = $app->$bs_prefix->begin;
+                $txn->db->select('channel',
+                    'id',
+                    { name => $CLEANUP_CHANNEL },
+                    { for => \'update skip locked' },
+                )->hash or return;
+                $txn->lock_state($CLEANUP_CHANNEL, sub ($state) {
+                    return undef, undef if Time::HiRes::time() < $state->{last_cleanup} + $cleanup_interval;
+
+                    $state->{last_cleanup} = $t0;
+                    $me_is_cleanup_executor = 1;
+
+                    return undef, $state;
                 });
                 return unless $me_is_cleanup_executor;
-
-                $app->$bs_prefix->global_cleanup;
+                await $app->$bs_prefix->global_cleanup;
+                $txn->commit;
             });
         }
     });
@@ -227,6 +237,10 @@ sub register ($self, $app, $config) {
     });
 
     $app->helper("$bs_prefix.create_channel_or_die", sub ($c, $channel_name, $state = undef, $attrs = {}) {
+        ! $c->$bs_prefix->db->select('channel',
+            'id',
+            { name => $channel_name },
+        )->hash or db_duplicate_error;
         my $type = $attrs->{type} // '';
         my $keep_events = exists $attrs->{keep_events} ? to_bool $attrs->{keep_events} : true;
         my $sth;
@@ -247,22 +261,16 @@ sub register ($self, $app, $config) {
     });
 
     $app->helper("$bs_prefix.create_channel", sub ($c, $channel_name, $state = undef, $attrs = {}) {
-        my $type = $attrs->{type} // '';
-        my $keep_events = exists $attrs->{keep_events} ? to_bool $attrs->{keep_events} : true;
-        my $db = $c->$bs_prefix->db;
-        ! $db->select('channel',
-            ['id'],
-            { name => $channel_name },
-        )->hash or return;
-        $db->insert('channel',
-            {
-                name        => $channel_name,
-                type        => $type,
-                state       => to_json($state),
-                keep_events => $keep_events,
-            },
-            { on_conflict => undef },
-        );
+        try {
+            $c->$bs_prefix->create_channel_or_die($channel_name, $state, $attrs);
+            return 1;
+        } catch {
+            my $e = $@;
+            if ($e->$_isa('BoardStreams::Exception::DbError::Duplicate')) {
+                return !!0;
+            }
+            die $e;
+        };
     });
 
     $app->helper("$bs_prefix.begin", sub ($c) {
@@ -364,6 +372,12 @@ sub register ($self, $app, $config) {
 
                 # join handler
                 my $join_handler = BoardStreams::Registry->get_join($channel_name) or return;
+                if (BoardStreams::Registry->has_pair($c, $channel_name)) {
+                    # not first join, so use a join handler that does nothing
+                    $join_handler = sub {
+                        return { limit => 0 };
+                    };
+                }
                 $c->$bs_prefix->channel_exists($channel_name) or return;
                 my $attrs = { since_id => $since_id, is_reconnect => int(defined $since_id) };
 
@@ -424,7 +438,7 @@ sub register ($self, $app, $config) {
                             data        => {
                                 event => from_json($event_row->{event}),
                             },
-                        }, "event_patch-$channel_name");
+                        }, "imm-event_patch-$channel_name-$channel_uuid");
                     }
                 }
 
@@ -434,12 +448,13 @@ sub register ($self, $app, $config) {
                     channelUUID => $channel_uuid,
                     id          => int $event_id,
                     data        => from_json($state),
-                }, "state-$channel_name");
+                }, "state-$channel_name-$channel_uuid");
             }
 
             # leave
             if ($hash->{type} eq 'leave') {
                 my $channel_name = string $hash->{channel};
+                return unless $c->$bs_prefix->joined($channel_name);
                 BoardStreams::Registry->remove_pair($c, $channel_name);
                 my $s = delete $c->stash->{subscriptions}{channels}{$channel_name};
                 $s->unsubscribe() if defined $s;
@@ -450,6 +465,7 @@ sub register ($self, $app, $config) {
             # action
             if ($hash->{type} eq 'action') {
                 my $channel_name = string $hash->{channel};
+                return unless $c->$bs_prefix->joined($channel_name);
                 my ($action_name, $payload) = $hash->{data}->@*;
                 my $action_sub = BoardStreams::Registry->get_action($channel_name, $action_name);
                 $action_sub->($channel_name, $c, $payload);
@@ -460,9 +476,13 @@ sub register ($self, $app, $config) {
                 my $channel_name = string $hash->{channel};
                 my ($request_name, $payload) = $hash->{data}->@*;
                 my $request_id = string $hash->{requestId};
-                my $request_sub = BoardStreams::Registry->get_request($channel_name, $request_name);
+                my $request_sub;
+                $c->$bs_prefix->joined($channel_name) or $request_sub //= sub {
+                    die { 'system' => 'not_joined' };
+                };
+                $request_sub //= BoardStreams::Registry->get_request($channel_name, $request_name);
                 $request_sub //= sub {
-                    die { system => 'invalid_request_name' };
+                    die { 'system' => 'invalid_request_name' };
                 };
                 my $t0 = Time::HiRes::time();
                 my $return_value;
@@ -543,7 +563,7 @@ sub register ($self, $app, $config) {
         };
     });
 
-    $app->helper("$bs_prefix.global_cleanup", sub ($c) {
+    $app->helper("$bs_prefix.global_cleanup", async sub ($c) {
         my $workers_state = $c->$bs_prefix->get_state($WORKERS_CHANNEL);
         my $time = get_time;
         foreach my $worker (values %$workers_state) {
@@ -561,7 +581,7 @@ sub register ($self, $app, $config) {
             # 1. find channel names from "guards" table
             my $worker_uuid = $worker->{uuid};
             my @channels = $db->query(q{
-                SELECT c.*
+                SELECT c.id, c.name
                     FROM guards g JOIN channel c ON (c.id = g.channel_id)
                     WHERE g.worker_uuid = ?
             }, $worker_uuid)->hashes->@*;
@@ -569,6 +589,7 @@ sub register ($self, $app, $config) {
             # 2. call cleanup subs (deleting guards rows in the process)
             CHANNEL:
             foreach my $channel (@channels) {
+                await next_tick_p();
                 my $txn = $c->$bs_prefix->begin;
                 my $db = $txn->db;
                 my $channel_name = $channel->{name};
@@ -578,7 +599,7 @@ sub register ($self, $app, $config) {
                     'id',
                     { name => $channel_name },
                     { for => 'update' },
-                );
+                )->hash or next CHANNEL;
                 my $guard_row = $db->select('guards',
                     ['counter'],
                     {
@@ -586,8 +607,7 @@ sub register ($self, $app, $config) {
                         channel_id  => $channel->{id},
                     },
                     { for => 'update' },
-                )->hash;
-                defined $guard_row or next CHANNEL;
+                )->hash or next CHANNEL;
                 my $guards_counter = $guard_row->{counter};
 
                 # call cleanup sub
@@ -690,7 +710,6 @@ sub register ($self, $app, $config) {
     {
         $app->$bs_prefix->create_channel($WORKERS_CHANNEL, {}, { keep_events => 0 });
         $app->$bs_prefix->create_channel($CLEANUP_CHANNEL, { last_cleanup => 0 }, { keep_events => 0 });
-        $app->$bs_prefix->global_cleanup();
     }
 }
 
